// Copyright (C) 1993 - Richard Barrett, Michael Berry, Tony F. Chan, James Demmel, June M. Donato, Jack Dongarra, Victor Eijkhout, Roldan Pozo, Charles Romine, and Henk Van der Vorst
// Copyright (C) 2005 - INRIA - Sage Group (IRISA)
// Copyright (C) 2011 - DIGITEO - Michael Baudin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

function [x, err, iter, flag, res]  = imsls_sor(varargin)
    // Solves linear equations using Successive Over-Relaxation Method.
    //
    // Calling Sequence
    // x  = imsls_sor(A, b)
    // x  = imsls_sor(A, b, x0)
    // x  = imsls_sor(A, b, x0, w)
    // x  = imsls_sor(A, b, x0, w, max_it)
    // x  = imsls_sor(A, b, x0, w, max_it, tol)
    // [x, err]  = imsls_sor(...)
    // [x, err, iter]  = imsls_sor(...)
    // [x, err, iter, flag]  = imsls_sor(...)
    // [x, err, iter, flag, res]  = imsls_sor(...)
    //
    // Parameters
    //    A:        a n-ny-n full or sparse matrix of doubles
    //    b:        a n-ny-1 full or sparse matrix of doubles, right hand side vector
    //    x0:       a n-ny-1 full or sparse matrix of doubles, initial guess vector  (default: zeros(n,1))
    //    w:        a 1-ny-1 matrix of doubles, relaxation scalar (default w=0.4)
    //    max_it:   a 1-ny-1 matrix of doubles, integer value,  maximum number of iterations (default n)
    //    tol:      a 1-ny-1 matrix of doubles, positive,  relative error tolerance on x (default 1000*%eps)
    //    x:        a n-ny-1 full or sparse matrix of doubles,  solution vector
    //    err:      a 1-ny-1 matrix of doubles, final relative residual norm. This is equal to norm(A*x-b)/norm(b) if norm(b) is nonzero, and is equal to norm(A*x-b) if norm(b) is zero.
    //    iter:     a 1-ny-1 matrix of doubles, integer value,  number of iterations performed
    //    flag:     a 1-ny-1 matrix of doubles, integer value, flag=0: solution found to tolerance, flag=1: no convergence given max_it
    //    res:      a (iter+1)-ny-1 full or sparse matrix of doubles, history of residual. res(1) is the initial residual and res(i+1) is the residual for the iteration i, for i=1,2,...,iter.
    //
    // Description
    // Solves the linear system Ax=b using the 
    // Successive Over-Relaxation Method (Gauss-Seidel method when omega = 1 ).
    //
    // Any optional input argument equal to the empty matrix [] is replaced by its default value.
	//
    // It is not possible to provide A as a function for imsls_jacobi.
	// This is because the solver internally splits the matrix A into M and N, then 
	// uses these matrices directly.
    //
    // Successive Overrelaxation (SOR) can be derived from the Gauss-Seidel method by
    // introducing an extrapolation parameter omega. 
    // For the optimal choice of omega, SOR may converge faster than Gauss-Seidel by an order of magnitude.
	//
	// If the coefficient matrix A is symmetric and positive definite, the SOR iteration 
	// is guaranteed to converge for any value of omega between 0 and 2, though the choice of omega 
	// can significantly affect the rate at which the SOR iteration converges.
    //
    // Examples
    // // No convergence
    // A=imsls_makefish(4); 
    // xe=(1:16)';
    // b=A*xe;
    // [x,err,iter,flag,res] = imsls_sor(A,b)
    //
    // // With initial guess
    // // No convergence
    // x0=zeros(16,1);
    // [x,err,iter,flag,res] = imsls_sor(A,b,x0);
    //
    // // Increase max_it: convergence
    // // Use default x0
    // w=0.8; 
    // max_it=200; 
    // tol=1000*%eps;
    // [x,err,iter,flag,res] = imsls_sor(A,b,[],w,max_it,tol);
    // 
    // Bibliography
    //     Univ. of Tennessee and Oak Ridge National Laboratory, October 1, 1993., Details of this algorithm are described in "Templates for the Solution of Linear Systems: Building Blocks for Iterative Methods", Barrett, Berry, Chan, Demmel, Donato, Dongarra, Eijkhout, Pozo, Romine, and van der Vorst, SIAM Publications, 1993. (ftp netlib2.cs.utk.edu; cd linalg; get templates.ps).
    //     "Templates for the Solution of Linear Systems: Building Blocks for Iterative Methods", Barrett, Berry, Chan, Demmel, Donato, Dongarra, Eijkhout, Pozo, Romine, and Van der Vorst, SIAM Publications, 1993 (ftp netlib2.cs.utk.edu; cd linalg; get templates.ps).
    //     "Iterative Methods for Sparse Linear Systems, Second Edition", Saad, SIAM Publications, 2003 (ftp ftp.cs.umn.edu; cd dept/users/saad/PS; get all_ps.zip).
    //     http://www.netlib.org/templates/matlab//sor.m
    //
    // Authors
    // Copyright (C) 1993 - Richard Barrett, Michael Berry, Tony F. Chan, James Demmel, June M. Donato, Jack Dongarra, Victor Eijkhout, Roldan Pozo, Charles Romine, and Henk Van der Vorst
    // Copyright (C) 2005 - INRIA - Sage Group (IRISA)
    // Copyright (C) 2011 - DIGITEO - Michael Baudin

    // Load Internals lib if it's not already loaded
    // =========================================================================
    if ~ exists("imslsinternalslib") then
        path = imsls_getpath()
        imslsinternalslib = lib(fullfile(path,"macros","internals"));
    end

    [lhs, rhs] = argn()
    apifun_checkrhs ( "imsls_sor" , rhs , 2:6 )
    apifun_checklhs ( "imsls_sor" , lhs , 0:5 )

    //
    // Get input arguments
    A = varargin(1);
    b = varargin(2);
    n = size(b,"*")
    x0 = apifun_argindefault(varargin , 3 , zeros(n,1) );
    w = apifun_argindefault(varargin , 4 , 0.4 );
    max_it = apifun_argindefault(varargin , 5 , n )
    tol = apifun_argindefault(varargin , 6 , 1000*%eps )
    //
    // Check types
    apifun_checktype ( "imsls_sor" , A ,  "A" , 1 , ["constant" "sparse"])
    apifun_checktype ( "imsls_sor" , b ,  "b" , 2 , ["constant" "sparse"])
    apifun_checktype ( "imsls_sor" , x0 , "x0" , 3 , ["constant" "sparse"])
    apifun_checktype ( "imsls_sor" , w ,  "w" , 4 , "constant")
    apifun_checktype ( "imsls_sor" , max_it , "max_it" , 5 , "constant")
    apifun_checktype ( "imsls_sor" , tol ,    "tol" , 6 , "constant")
    //
    // Check size
    if ( or ( typeof(A) == ["constant" "sparse"]) ) then
        apifun_checksquare ( "imsls_sor" , A , "A" , 1 )
    end
    apifun_checkveccol ( "imsls_sor" , b , "b" , 2 , size(b,"*") )
    apifun_checkveccol ( "imsls_sor" , x0 , "x0" , 3 , size(x0,"*") )
    apifun_checkscalar ( "imsls_sor" , w , "w" , 4 )
    apifun_checkscalar ( "imsls_sor" , max_it , "max_it" , 5 )
    apifun_checkscalar ( "imsls_sor" , tol , "tol" , 6 )
    //
    // Check content
    apifun_checkgreq ( "imsls_sor" , max_it , "max_it" , 5 , 1 )
	apifun_checkflint( "imsls_sor" , max_it , "max_it" , 5)
    apifun_checkgreq ( "imsls_sor" , tol , "tol" , 6 , number_properties("tiny") )

	// initialization
    flag = 0;                                   
	iter = 0;
    i = 0;
    x = x0;

    bnrm2 = norm( b );
    if  ( bnrm2 == 0.0 ) then 
        bnrm2 = 1.0; 
    end

    r = b - A*x;
    err = norm( r ) / bnrm2;
    res = err;
    if ( err < tol ) then
        return; 
    end

    [ M, N, b ] = imsls_split( A, b, w, 2 );          // matrix splitting

    for i = 1:max_it                         // begin iteration

        x_1 = x;
        x   = M \ ( N*x + b );                   // update approximation

        err = norm( x - x_1 ) / norm( x );     // compute error
        res = [res;err];
        if ( err <= tol ) then
            iter=i; 
            break; 
        end          // check convergence

        if ( i == max_it ) then 
            iter=i; 
        end
    end
    b = b / w;                                  // restore rhs

    if ( err > tol ) then 
        flag = 1; 
    end;           // no convergence

    if ( flag <> 0 ) then
        warning(msprintf(gettext("%s: Algorithm has not converged."),"imsls_sor"))
    end
endfunction
